@@ -1,12 +1,14 @@
 import base64
-import os
 import logging
-import requests
+import os
 import sched
-import time
 import threading
-from flask import request as flaskrequest
+import time
+
+import requests
 from flask import Flask, jsonify
+from flask import request as flaskrequest
+from requests_oauthlib import OAuth2Session
 from werkzeug.local import Local
 
 from .flowrun import FlowRun
@@ -39,6 +41,12 @@ def _start_step(name, operation, step):
         return jsonify(**{})
 
 
+@app.route("/<instance>/token", methods=["GET"])
+def _token_validator(instance):
+    slims_instances[instance].handle_oauth_code(flaskrequest.args.get('code'))
+    return 'Python instance "' + instance + '" registered'
+
+
 def _flask_thread(port):
     app.run(port=port, host='0.0.0.0')
 
@@ -49,20 +57,31 @@ class _SlimsApiException(Exception):
 
 class _SlimsApi(object):
 
-    def __init__(self, url, username, password, repo_location):
+    def __init__(self, url, username, password, repo_location, oauth=False, token_updater=None, redirect_url=""):
         self.url = url + "/rest/"
+        self.raw_url = url + "/"
         self.username = username
         self.password = password
         self.repo_location = repo_location
+        self.oauth = oauth
+        if oauth:
+            self.oauth_session = OAuth2Session("python-remote",
+                                               redirect_uri=redirect_url,
+                                               scope="api",
+                                               auto_refresh_url=self.raw_url + "oauth/token",
+                                               token_updater=token_updater)
 
     def get_entities(self, url, body=None):
         if not url.startswith(self.url):
             url = self.url + url
 
-        response = requests.get(url,
-                                auth=(self.username, self.password),
-                                headers=_SlimsApi._headers(),
-                                json=body)
+        if self.oauth:
+            response = self.oauth_session.get(url, headers=_SlimsApi._headers(), json=body)
+        else:
+            response = requests.get(url,
+                                    auth=(self.username, self.password),
+                                    headers=_SlimsApi._headers(),
+                                    json=body)
         records = []
         if response.status_code == 200:
             for entity in response.json()["entities"]:
@@ -75,20 +94,38 @@ class _SlimsApi(object):
             raise _SlimsApiException("Could not fetch entities: " + response.text)
 
     def get(self, url):
-        return requests.get(self.url + url,
-                            auth=(self.username, self.password), headers=_SlimsApi._headers())
+        if self.oauth:
+            return self.oauth_session.get(self.url + url, headers=_SlimsApi._headers())
+        else:
+            return requests.get(self.url + url,
+                                auth=(self.username, self.password), headers=_SlimsApi._headers())
 
     def post(self, url, body):
-        return requests.post(self.url + url, json=body,
-                             auth=(self.username, self.password), headers=_SlimsApi._headers())
+        if self.oauth:
+            return self.oauth_session.post(self.url + url, json=body, headers=_SlimsApi._headers())
+        else:
+            return requests.post(self.url + url, json=body,
+                                 auth=(self.username, self.password), headers=_SlimsApi._headers())
 
     def put(self, url, body):
-        return requests.put(self.url + url, json=body,
-                            auth=(self.username, self.password), headers=_SlimsApi._headers())
+        if self.oauth:
+            return self.oauth_session.put(self.url + url, json=body, headers=_SlimsApi._headers())
+        else:
+            return requests.put(self.url + url, json=body,
+                                auth=(self.username, self.password), headers=_SlimsApi._headers())
 
     def delete(self, url):
-        return requests.delete(self.url + url,
-                               auth=(self.username, self.password), headers=_SlimsApi._headers())
+        if self.oauth:
+            return self.oauth_session.delete(self.url + url, headers=_SlimsApi._headers())
+        else:
+            return requests.delete(self.url + url,
+                                   auth=(self.username, self.password), headers=_SlimsApi._headers())
+
+    def authorization_url(self):
+        return self.oauth_session.authorization_url(self.raw_url + "oauth/authorize")[0]
+
+    def fetch_token(self, code):
+        return self.oauth_session.fetch_token(self.raw_url + 'oauth/token', code=code)
 
     @staticmethod
     def _headers():
@@ -107,7 +144,7 @@ class Slims(object):
         url (str): The url of the REST API of this slims instance
         username (str, optional): The username to login with (needed for standard operations)
         password (str, optional): The password to login with (needed for standard operations)
-        token (str, optional): The token to login with (needed to add SLimsGate flows)
+        oauth (bool, optional): Whether Oauth authentication is used
         repo_location (str, optional): The location of the file repository (this can
             be used to access attachments without needing to download them)
         local_host (str, optional): The IP on which this python script is running
@@ -123,28 +160,35 @@ class Slims(object):
                  url,
                  username=None,
                  password=None,
-                 token=None,
+                 oauth=False,
                  repo_location=None,
                  local_host="localhost",
                  local_port=5000):
 
         slims_instances[name] = self
+        self.local_host = local_host
+        self.local_port = local_port
+        self.local_url = "http://" + self.local_host + ":" + str(self.local_port) + "/"
         if username is not None and password is not None:
             self.slims_api = _SlimsApi(url, username, password, repo_location)
-        elif token is not None:
-            self.slims_api = _SlimsApi(url, "TOKEN", token, repo_location)
+        elif oauth:
+            self.slims_api = _SlimsApi(url, "OAUTH", "oauth", repo_location, True, self.token_updater,
+                                       self.local_url + name + "/token")
+            self.token = None
         else:
-            raise Exception("Either specify a username and a password or a token")
+            raise Exception("Either specify a username and a password or use oauth")
 
         self.name = name
+        self.oauth = oauth
         self.operations = {}
         self.flow_definitions = []
         self.refresh_flows_thread = threading.Thread(target=self._refresh_flows_thread_inner)
         self.refresh_flows_thread.daemon = True
-        self.local_host = local_host
-        self.local_port = local_port
 
-    def fetch(self, table, criteria, sort=[], start=None, end=None):
+    def token_updater(self, token):
+        self.token = token
+
+    def fetch(self, table, criteria, sort=None, start=None, end=None):
         """Fetch data by criteria
 
         The optional start and end parameters can be used to page the returned
@@ -178,6 +222,8 @@ class Slims(object):
             Fetches content records that have an id that starts with DNA. The
             returned list is sorted by cntn_barCode (descending).
         """
+        if sort is None:
+            sort = []
         body = {
             "sortBy": sort,
             "startRow": start,
@@ -233,7 +279,7 @@ class Slims(object):
         new_values = response.json()["entities"][0]
         return Record(new_values, self.slims_api)
 
-    def add_flow(self, flow_id, name, usage, steps, testing=False, last_flow=True):
+    def add_flow(self, flow_id, name, usage, steps, testing=False):
         """Add a new SLimsGate flow to the slims interface
 
         Note:
@@ -245,7 +291,7 @@ class Slims(object):
             name(string): Displayed name of the the flow
             usage(string): Usage of the slimsgate flow
             steps(list step): The steps of the slimsgate flow
-            last_flow(boolean): Defines if this is the last flow you will add (Default True)
+            testing(bool): Dry run
 
         Examples:
             >>> def hello_world(flow_run):
@@ -260,6 +306,10 @@ class Slims(object):
                             action=hello_world
                         )])
         """
+
+        if not self.oauth:
+            raise Exception('Python flows require oauth to be enabled')
+
         step_dicts = []
         i = 0
         for step in steps:
@@ -270,13 +320,16 @@ class Slims(object):
 
         flow = {'id': flow_id, 'name': name, 'usage': usage, 'steps': step_dicts, 'pythonApiFlow': True}
         self.flow_definitions.append(flow)
-        self._register_flows(self.flow_definitions, False)
+        if self.token is None and not testing:
+            print("Visit " + self.slims_api.authorization_url())
+            _flask_thread(self.local_port)
+        else:
+            self._register_flows([flow], False)
 
-        if not testing:
-            if not self.refresh_flows_thread.is_alive():
-                self.refresh_flows_thread.start()
-            if last_flow:
-                _flask_thread(self.local_port)
+    def handle_oauth_code(self, code):
+        self.token = self.slims_api.fetch_token(code)
+        if not self.refresh_flows_thread.is_alive():
+            self.refresh_flows_thread.start()
 
     def _register_flows(self, flows, is_reregister):
         flow_ids = map(lambda flow: flow.get('id'), flows)
@@ -291,7 +344,7 @@ class Slims(object):
                 logger.info("Successfully " + verb + "ed " + str(flow_ids))
             else:
                 logger.info("Could not " + verb + " " + str(flow_ids) +
-                            + " (HTTP Response code: " + str(response.status_code) + ")")
+                            " (HTTP Response code: " + str(response.status_code) + ")")
                 try:
                     logger.info("Reason: " + response.json()["errorMessage"])
                 except Exception:
@@ -306,12 +359,12 @@ class Slims(object):
         return output
 
     def _refresh_flows_thread_inner(self):
-        def refresh_flows(scheduler):
+        def refresh_flows(internal_scheduler):
             self._register_flows(self.flow_definitions, True)
-            scheduler.enter(60, 1, refresh_flows, (scheduler,))
+            internal_scheduler.enter(60, 1, refresh_flows, (internal_scheduler,))
 
         scheduler = sched.scheduler(time.time, time.sleep)
-        scheduler.enter(60, 1, refresh_flows, (scheduler,))
+        scheduler.enter(5, 1, refresh_flows, (scheduler,))
         scheduler.run()
 
 
